// ------------------------------------------------------------
QUnit.module('Matrix internal module', {
  before: function() {
    // Generate a random non square matrix
	 // Generate the dimensions
	var max = 10;
	var min = 2;
	var r = Math.floor(Math.random()*(max-min+1) + min);
	var c = Math.floor(Math.random()*(max-min+1) + min);
	while (c == r) {
	  c = Math.floor(Math.random()*(max-min+1) + min);
	}
	
	// Generate the values 
	var doubleArray = new Array(r);
    for (var i = 0; i < r; ++i) {
	  doubleArray[i] = new Array(c);
	  for (var j = 0; j < c; ++j) {
	    doubleArray[i][j] = Math.random();
	  }
    }
	this.nsMatValues = doubleArray;	
	
	// Generate a random square matrix
	var doubleArray2 = new Array(r);
    for (var i = 0; i < r; ++i) {
	  doubleArray2[i] = new Array(r);
	  for (var j = 0; j < r; ++j) {
	    doubleArray2[i][j] = Math.random();
	  }
    }
	this.sMatValues = doubleArray2;	
  }
});


QUnit.test('Matrix basic manipulations', function(assert) {    
  // Test non square matrix using random data
  {
      var nsMat = new PortfolioAllocation.Matrix(this.nsMatValues);
    
      // Check the dimensions	
      assert.equal(nsMat.nbRows, this.nsMatValues.length, 'Number of rows');
      assert.equal(nsMat.nbColumns, this.nsMatValues[0].length, 'Number of columns');
      
      // Ensure the matrix is not square
      assert.equal(nsMat.isSquare(), false, 'Matrix not square');

      // Ensure the matrix is not symmetric
      assert.equal(nsMat.isSymmetric(), false, 'Matrix not symmetric');
	  
      // Ensure the matrix is equal to itself
      assert.equal(PortfolioAllocation.Matrix.areEqual(nsMat,nsMat), true, 'Matrix equal to itself');
      
      // Check the values with getValue(At)
      for (var i = 0; i < nsMat.nbRows; ++i) {
    	for (var j = 0; j < nsMat.nbColumns; ++j) {
    	  assert.equal(nsMat.getValueAt(i+1,j+1), this.nsMatValues[i][j], 'Matrix values');
		  assert.equal(nsMat.getValue(i+1,j+1), this.nsMatValues[i][j], 'Matrix values');
    	}
      }
  }
  
  // Test equality using the non square matrix
  {
      // Create a new matrix equal to the previous one and replace 
      // the values of the matrix with new values with setValue(At)
      var nsMat2 = new PortfolioAllocation.Matrix(this.nsMatValues);
      for (var i = 1; i <= nsMat.nbRows; ++i) {
    	for (var j = 1; j <= nsMat.nbColumns; ++j) {
    	  var newVal = nsMat.getValueAt(i,j) + Math.random();
    	  nsMat2.setValueAt(i,j, newVal);
    	  assert.equal(nsMat2.getValueAt(i,j), newVal, 'Matrix values #2 - 1/2');
		  
    	  var newVal2 = nsMat.getValueAt(i,j) + Math.random();
    	  nsMat2.setValue(i,j, newVal2);
    	  assert.equal(nsMat2.getValueAt(i,j), newVal2, 'Matrix values #2 - 2/2');
    	}
      }
    
      // Ensure the old matrix is not equal to the new matrix with a 0 tolerance...
      assert.equal(PortfolioAllocation.Matrix.areEqual(nsMat, nsMat2), false, 'Old matrix not strictly equal to new matrix');
    
      // ... but is with a 1 tolerance, as Math.random max value is 1, so that the new matrix coefficients must be 
      // equal to the old matrix coefficients + at most 1 !
      assert.equal(PortfolioAllocation.Matrix.areEqual(nsMat, nsMat2, 1), true, 'Old matrix equal to new matrix within 1');
  }
  
  // Test square matrix using random data
  {
      var sMat = new PortfolioAllocation.Matrix(this.sMatValues);
    
      // Ensure the matrix is square
      assert.equal(sMat.isSquare(), true, 'Matrix is square');
      
      // Extract the matrix diagonal elements
      var diag = sMat.diagonal();
      for (var i = 1; i <= sMat.nbRows; ++i) {
          assert.equal(diag.getValueAt(i, 1), sMat.getValueAt(i, i), 'Matrix diagonal elements');
      }
  }
  
  // Test symmetric matrix using random data
  {
	  // Create a square matrix
	  var sMat = new PortfolioAllocation.Matrix(this.sMatValues);
	  
	  // Perturbate it to ensure it is not symmetric
	  sMat.setValueAt(1,2, 1);
	  sMat.setValueAt(2,1, -1);
	  assert.equal(sMat.isSymmetric(), false, 'Matrix not symmetric');
	  
	  // Multiply it with its transpose, which ensures the resulting matrix is symmetric
	  var symMat = PortfolioAllocation.Matrix.txy(sMat, sMat);
	  assert.equal(sMat.isSymmetric(), false, 'Matrix symmetric');
	  
	  // Perturbate it to ensure it is numerically symmetric, depending on the precision
	  symMat.setValueAt(1,2, 1);
	  symMat.setValueAt(2,1, 1.01);
	  assert.equal(symMat.isSymmetric(), false, 'Matrix not numerically symmetric');
	  assert.equal(symMat.isSymmetric(1e-2), false, 'Matrix not numerically symmetric');
	  assert.equal(symMat.isSymmetric(1e-1), true, 'Matrix numerically symmetric');
  }

});

QUnit.test('Matrix-matrix product', function(assert) {    
  // Test matrix-matrix product using static data
  var mat1 = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
  var mat2 = new PortfolioAllocation.Matrix([[1,2], [2,4], [3,6]]);

  var prodMat = PortfolioAllocation.Matrix.xy(mat1, mat2);
  var expectedResMat = new PortfolioAllocation.Matrix([[14,28], [32,64]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(prodMat, expectedResMat), true, 'Matrix-matrix XY');

  var axyMat = PortfolioAllocation.Matrix.axy(2, mat1, mat2);
  var expectedResMat = new PortfolioAllocation.Matrix([[28,56], [64, 128]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(axyMat, expectedResMat), true, 'Matrix-matrix AXY');

  var mat3 = mat1.transpose();
  var atxyMat = PortfolioAllocation.Matrix.atxy(2, mat3, mat2);
  assert.equal(PortfolioAllocation.Matrix.areEqual(atxyMat, expectedResMat), true, 'Matrix-matrix AX^tY');

  var mat4 = mat2.transpose();
  var axtyMat = PortfolioAllocation.Matrix.axty(2, mat1, mat4);
  assert.equal(PortfolioAllocation.Matrix.areEqual(axtyMat, expectedResMat), true, 'Matrix-matrix AXY^t');
  
  // TODO: use random data
});


QUnit.test('Matrix-matrix elementwise product', function(assert) {    
  // Test matrix-matrix elementwise product using static data
  var mat1 = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
  var mat2 = new PortfolioAllocation.Matrix([[1,2,3]]);
  var mat3 = new PortfolioAllocation.Matrix([[1],[4]]);
  var mat4 = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);

  var prodMat = PortfolioAllocation.Matrix.elementwiseProduct(mat1, mat2);
  var expectedResMat = new PortfolioAllocation.Matrix([[1,4,9], [4,10,18]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(prodMat, expectedResMat), true, 'Matrix-matrix elementwise product - Row matrix');

  var prodMat = PortfolioAllocation.Matrix.elementwiseProduct(mat1, mat3);
  var expectedResMat = new PortfolioAllocation.Matrix([[1,2,3], [16,20,24]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(prodMat, expectedResMat), true, 'Matrix-matrix elementwise product - Column matrix');

  var prodMat = PortfolioAllocation.Matrix.elementwiseProduct(mat1, mat4);
  var expectedResMat = new PortfolioAllocation.Matrix([[1,4,9], [16,25,36]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(prodMat, expectedResMat), true, 'Matrix-matrix elementwise product - Full matrix');

  // TODO: use random data
});


QUnit.test('Matrix rows and columns swaps', function(assert) {    
	// Test with static data
	//
	// Swap the same row
	{
		var mat = new PortfolioAllocation.Matrix([[1,2], [4,5]]);

		var resMatCopy = mat.swapRows(1,1);
		var resMatInPlace = mat.swapRows(1,1, {inPlace: true});
		var expectedMat = new PortfolioAllocation.Matrix(mat);
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatCopy, expectedMat), true, 'Swap the same row, copy');
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatInPlace, expectedMat), true, 'Swap the same row, in place');

		mat.touch = true;
		assert.equal(resMatCopy.touch, undefined, 'Swap the same row, copy, original matrix not altered');
		assert.equal(resMatInPlace.touch, true, 'Swap the same row, in place, original matrix altered');
	} 

	// Test with static data
	//
	// Swap two rows
	{
		var mat = new PortfolioAllocation.Matrix([[1,2], [4,5]]);

		var resMatCopy = mat.swapRows(1,2);
		var resMatInPlace = mat.swapRows(1,2, {inPlace: true});
		var expectedMat = new PortfolioAllocation.Matrix([[4,5], [1,2]]);
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatCopy, expectedMat), true, 'Swap two rows, copy');
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatInPlace, expectedMat), true, 'Swap two rows, in place');

		mat.touch = true;
		assert.equal(resMatCopy.touch, undefined, 'Swap two rows, copy, original matrix not altered');
		assert.equal(resMatInPlace.touch, true, 'Swap two rows, in place, original matrix altered');
	} 
	
	
	// Test with static data
	//
	// Swap the same column
	{
		var mat = new PortfolioAllocation.Matrix([[1,2], [4,5]]);

		var resMatCopy = mat.swapColumns(1,1);
		var resMatInPlace = mat.swapColumns(1,1, {inPlace: true});
		var expectedMat = new PortfolioAllocation.Matrix(mat);
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatCopy, expectedMat), true, 'Swap the same row, copy');
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatInPlace, expectedMat), true, 'Swap the same row, in place');

		mat.touch = true;
		assert.equal(resMatCopy.touch, undefined, 'Swap the same row, copy, original matrix not altered');
		assert.equal(resMatInPlace.touch, true, 'Swap the same row, in place, original matrix altered');
	} 

	// Test with static data
	//
	// Swap two columns
	{
		var mat = new PortfolioAllocation.Matrix([[1,2], [4,5]]);

		var resMatCopy = mat.swapColumns(1,2);
		var resMatInPlace = mat.swapColumns(1,2, {inPlace: true});
		var expectedMat = new PortfolioAllocation.Matrix([[2,1], [5,4]]);
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatCopy, expectedMat), true, 'Swap two columns, copy');
		assert.equal(PortfolioAllocation.Matrix.areEqual(resMatInPlace, expectedMat), true, 'Swap two columns, in place');

		mat.touch = true;
		assert.equal(resMatCopy.touch, undefined, 'Swap two columns, copy, original matrix not altered');
		assert.equal(resMatInPlace.touch, true, 'Swap two columns, in place, original matrix altered');
	}

  // TODO: use random data
});


QUnit.test('LU decomposition', function(assert) {    
	// Test with static data
	{
		var mat = new PortfolioAllocation.Matrix([[3,17, 10], [2, 4, -2], [6, 18, -12]]);

		// Test the matrix format for the permutation matrices
		var expectedL = new PortfolioAllocation.Matrix([[1, 0, 0], [0.9444444444444444, 1, 0], [0.2222222222222222, 0.031249999999999993, 1]]);
		var expectedU = new PortfolioAllocation.Matrix([[18, -12, 6], [0, 21.333333333333332, -2.666666666666666], [0, 0, 0.75]]);
		var expectedP = new PortfolioAllocation.Matrix([[0, 0, 1], [1 ,0 ,0], [0, 1, 0]]);
		var expectedQ = new PortfolioAllocation.Matrix([[0, 0, 1], [1 ,0 ,0], [0, 1, 0]]);

		var lu = PortfolioAllocation.Matrix.luDecomposition(mat);
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.lowerTriangular, expectedL), true, 'Test #1, L');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.upperTriangular, expectedU), true, 'Test #1, U');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.rowPermutation, expectedP), true, 'Test #1, P');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.columnPermutation, expectedQ), true, 'Test #1, Q');

		var mlu = PortfolioAllocation.Matrix.xy(lu.lowerTriangular, lu.upperTriangular);
		var mpaq = PortfolioAllocation.Matrix.xy(lu.rowPermutation, PortfolioAllocation.Matrix.xy(mat, lu.columnPermutation));
		assert.equal(PortfolioAllocation.Matrix.areEqual(mlu, mpaq), true, 'Test #1, PAQ = LU');

		
		// Test the vector format for the permutation matrices
		var expectedPvector = new PortfolioAllocation.Matrix([3, 1, 2]);
		var expectedQvector = new PortfolioAllocation.Matrix([2, 3, 1]);

		var lu = PortfolioAllocation.Matrix.luDecomposition(mat, {permutationsOutputForm: "vector"});
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.lowerTriangular, expectedL), true, 'Test #1, L vector');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.upperTriangular, expectedU), true, 'Test #1, U vector');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.rowPermutation, expectedPvector), true, 'Test #1, P vector');
		assert.equal(PortfolioAllocation.Matrix.areEqual(lu.columnPermutation, expectedQvector), true, 'Test #1, Q vector');
	}

	// TODO: Test with random data
});



QUnit.test('Matrix element wise function', function(assert) {    
  // Test matrix element wise function using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
  var powerMat = mat.elemMap(function(i,j,val) { return Math.pow(val, 2); });
 
  var expectedResMat = new PortfolioAllocation.Matrix([[1,4,9], [16,25,36]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(powerMat, expectedResMat), true, 'Matrix element wise function');
});


QUnit.test('Matrix to string', function(assert) {    
  // Test matrix to string using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,10]]);
  assert.equal(mat.toString(), '[  1  2  3 ]\n[  4  5 10 ]\n', 'Matrix to string');
});


QUnit.test('Marix negativity and positivity functions', function(assert) {    
  // Test using static data
  {
	  var nothingMat = new PortfolioAllocation.Matrix([[-1,2,3], [4,5,10]]);
	  var posMat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,10]]);
	  var negMat = new PortfolioAllocation.Matrix([[-1,-2,-3], [-4,-5,-10]]);
	  var nonNegMat = new PortfolioAllocation.Matrix([[0,2,3], [4,5,10]]);
	  var nonPosMat = new PortfolioAllocation.Matrix([[0,-2,-3], [-4,-5,-10]]);
	  
	  assert.equal(posMat.isNonNegative(), true, 'Marix negativity and positivity - Test #1/1');
	  assert.equal(posMat.isPositive(), true, 'Marix negativity and positivity - Test #1/2');
	  assert.equal(posMat.isNonPositive(), false, 'Marix negativity and positivity - Test #1/3');
	  assert.equal(posMat.isNegative(), false, 'Marix negativity and positivity - Test #1/4');
	  
	  assert.equal(nonNegMat.isNonNegative(), true, 'Marix negativity and positivity - Test #2/1');
	  assert.equal(nonNegMat.isPositive(), false, 'Marix negativity and positivity - Test #2/2');
	  assert.equal(nonNegMat.isNonPositive(), false, 'Marix negativity and positivity - Test #2/3');
	  assert.equal(nonNegMat.isNegative(), false, 'Marix negativity and positivity - Test #2/4');  
	  
	  assert.equal(nothingMat.isNonNegative(), false, 'Marix negativity and positivity - Test #3/1');
	  assert.equal(nothingMat.isPositive(), false, 'Marix negativity and positivity - Test #3/2');
	  assert.equal(nothingMat.isNonPositive(), false, 'Marix negativity and positivity - Test #3/3');
	  assert.equal(nothingMat.isNegative(), false, 'Marix negativity and positivity - Test #3/4');
	  
	  assert.equal(negMat.isNonNegative(), false, 'Marix negativity and positivity - Test #4/1');
	  assert.equal(negMat.isPositive(), false, 'Marix negativity and positivity - Test #4/2');
	  assert.equal(negMat.isNonPositive(), true, 'Marix negativity and positivity - Test #4/3');
	  assert.equal(negMat.isNegative(), true, 'Marix negativity and positivity - Test #4/4');

	  assert.equal(nonPosMat.isNonNegative(), false, 'Marix negativity and positivity - Test #5/1');
	  assert.equal(nonPosMat.isPositive(), false, 'Marix negativity and positivity - Test #5/2');
	  assert.equal(nonPosMat.isNonPositive(), true, 'Marix negativity and positivity - Test #5/3');
	  assert.equal(nonPosMat.isNegative(), false, 'Marix negativity and positivity - Test #5/4');
  }

});


QUnit.test('Matrix to row array', function(assert) {    
  // Test using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,10]]);
  assert.deepEqual(mat.toRowArray(), [[1,2,3], [4,5,10]], 'Matrix to double array');
  assert.deepEqual(mat.toRowArray(function(i,j,val) { return i==j; }), [[1], [5]], 'Matrix to doule array with function');
  
  // Test using the random matrix, using identity matrix(matrix to array) == matrix
  var nsMat = new PortfolioAllocation.Matrix(this.nsMatValues);
  var nsMatArray = nsMat.toRowArray();
  assert.deepEqual(nsMatArray, this.nsMatValues, 'Array to matrix to array');
  assert.equal(PortfolioAllocation.Matrix.areEqual(nsMat, new PortfolioAllocation.Matrix(nsMatArray)), true, 'Matrix to double array to matrix');
});


QUnit.test('Matrix to array', function(assert) {    
  // Test using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,10]]);
  assert.deepEqual(mat.toArray(), [1,2,3,4,5,10], 'Matrix to array');
  assert.deepEqual(mat.toArray(function(i,j,val) { return i==j; }), [1, 5], 'Matrix to array with function');  
});


QUnit.test('Diagonal matrix creation', function(assert) {    
  // Test using static data
  var mat = PortfolioAllocation.Matrix.diagonal(new PortfolioAllocation.Matrix([1,2,3]));
  var expectedMat = new PortfolioAllocation.Matrix([[1,0,0], [0,2,0], [0,0,3]]);
  assert.deepEqual(mat.toArray(), expectedMat.toArray(), 'Diagonal matrix creation');
});


QUnit.test('Symetric matrix creation', function(assert) {    
  // Test using static data
  var mat = PortfolioAllocation.Matrix.fillSymmetric(2, function(i,j) { return i+j; });
  var expectedMat = new PortfolioAllocation.Matrix([[2,3], [3,4]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat, expectedMat), true, 'Symetric matrix creation');
});


QUnit.test('Submatrix extraction', function(assert) {    
  // Test using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6], [7,8,9]]);
  var subMat = mat.submatrix([1,3], [1, 3]);
  var expectedMat = new PortfolioAllocation.Matrix([[1,3], [7,9]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(subMat, expectedMat), true, 'Submatrix extraction');
});

QUnit.test('Matrix rows and columns extraction', function(assert) {    
  // Test using static data
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
  var expected_row_1 =  new PortfolioAllocation.Matrix([1,2,3]);
  var expected_row_2 =  new PortfolioAllocation.Matrix([4,5,6]);
  var expected_col_1 =  new PortfolioAllocation.Matrix([1,4]);
  var expected_col_2 =  new PortfolioAllocation.Matrix([2,5]);
  var expected_col_3 =  new PortfolioAllocation.Matrix([3,6]);
  
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat.row(1), expected_row_1), true, 'Matrix row extraction #1');
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat.row(2), expected_row_2), true, 'Matrix row extraction #2');
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat.column(1), expected_col_1), true, 'Matrix column extraction #1');
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat.column(2), expected_col_2), true, 'Matrix column extraction #2');
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat.column(3), expected_col_3), true, 'Matrix column extraction #3');
});

QUnit.test('Zeros matrix creation', function(assert) {    
  // Test using static data
  var mat = PortfolioAllocation.Matrix.zeros(3, 2);
  var expectedMat = new PortfolioAllocation.Matrix([[0,0], [0,0], [0,0]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat, expectedMat), true, 'Zeros matrix creation');
});

QUnit.test('Ones matrix creation', function(assert) {    
  // Test using static data
  var mat = PortfolioAllocation.Matrix.ones(3, 2);
  var expectedMat = new PortfolioAllocation.Matrix([[1,1], [1,1], [1,1]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat, expectedMat), true, 'Ones matrix creation');
});

QUnit.test('Identity matrix creation', function(assert) {    
  // Test using static data
  var mat = PortfolioAllocation.Matrix.identity(3);
  var expectedMat = new PortfolioAllocation.Matrix([[1,0,0], [0,1,0], [0,0,1]]);
  assert.equal(PortfolioAllocation.Matrix.areEqual(mat, expectedMat), true, 'Identity matrix creation');
});

QUnit.test('Transpose matrix', function(assert) {    
  // Test using static data  
  var mat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
  var transpMat = mat.transpose();
  var expectedMat = new PortfolioAllocation.Matrix([[1,4], [2,5], [3,6]]); 
  assert.equal(PortfolioAllocation.Matrix.areEqual(transpMat, expectedMat), true, 'Transpose matrix');
});


QUnit.test('QR decomposition', function(assert) {    
  // Test using static data  
  {
	  var mat = new PortfolioAllocation.Matrix([[1,2,0], [1,1,1], [2,1,0]]);
	  
	  // Computation of a QR decomposition
	  var fullQR = PortfolioAllocation.Matrix.qrDecomposition(mat);
	  var Q = fullQR[0];
	  var R = fullQR[1];
	  
	  var expectedQ = new PortfolioAllocation.Matrix([[0.408248290463863, 0.8616404368553292, -0.30151134457776363], 
													  [0.408248290463863, 0.12309149097933275, 0.9045340337332908], 
													  [0.816496580927726, -0.4923659639173309, -0.30151134457776363]]); 
	  var expectedR = new PortfolioAllocation.Matrix([[2.449489742783178, 2.041241452319315, 0.408248290463863], 
													  [0, 1.3540064007726602, 0.12309149097933275], 
													  [0, 0, 0.9045340337332908]]);
	  
	  assert.equal(PortfolioAllocation.Matrix.areEqual(Q, expectedQ), true, 'QR decomposition, full - #1 1/2');
	  assert.equal(PortfolioAllocation.Matrix.areEqual(R, expectedR), true, 'QR decomposition, full - #1 2/2');
	  
	  // Computation of the same QR decomposition, without Q
	  var qLessR = PortfolioAllocation.Matrix.qrDecomposition(mat, {qLess: true});
	  assert.equal(PortfolioAllocation.Matrix.areEqual(qLessR, R), true, 'QR decomposition, Q less - #2');
  }
  
  // TODO: Test using random data: check Q,R dimensions, check Q*R = A, check R upper triangular, check Q orthogonal: Q*Q^t = Identity (m)
  
  // TODO: Test error case
});

QUnit.test('Cholesky decomposition', function(assert) {    
  // Test using static data  
  {
	  // Computation of the Cholesky decomposition, which must fail because the matrix is not positive definite
	  var mat = new PortfolioAllocation.Matrix([[0.005, 0.897, 0.698], [0.897, 1.003, 0.302], [0.698, 0.302, 1.001]]);
	  assert.throws(function() { PortfolioAllocation.Matrix.choleskyDecomposition(mat); },
						 new Error('input matrix must be positive definite'),
						 "Cholesky decomposition - Non definite positive");

	  // Computation of the Cholesky decomposition
	  var mat = new PortfolioAllocation.Matrix([[1.005, 0.897, 0.698], [0.897, 1.003, 0.302], [0.698, 0.302, 1.001]]);
	  var chol = PortfolioAllocation.Matrix.choleskyDecomposition(mat);
	  var G = chol.lowerTriangular;
	  
	  // Ensure G is lower triangular with positive diagonal elements
	  var lowerTriangular = true;
	  var positiveDiagonalElements = true;
	  for (var i = 1; i <= 3; ++i) {
		  if (G.getValueAt(i,i) <= 0) {
			  positiveDiagonalElements = false;
		  }
		  
		  for (var j = i+1; j <= 3; ++j) {
			  if (G.getValueAt(i,j) != 0) {
				  lowerTriangular = false;
			  }
		  }
	  }
	  assert.equal(lowerTriangular, true, 'Cholesky decomposition - Lower triangular');
	  assert.equal(positiveDiagonalElements, true, 'Cholesky decomposition - Positive diagonal elements');
	  
	  // Ensure A = G*G^t
	  var mmat = PortfolioAllocation.Matrix.axty(1, G, G);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(mat, mmat), true, 'Cholesky decomposition - A = G*G^t');
						 
  }
  
	// Test using static data that semi-definite positive matrices are NOT managed with a standard
	// Cholesky decomposition.
	{
		var mat = new PortfolioAllocation.Matrix([[1, -1, 1], [-1, 1, -1], [1, -1, 2]]);
		assert.throws(function() { 
			var chol = PortfolioAllocation.Matrix.choleskyDecomposition(mat); 
		},
		new Error('input matrix must be positive definite'),
		"Cholesky decomposition - semi-definite positive KO");
  }
  
	// Test using static data that semi-definite positive matrices are properly managed with a Cholesky decomposition
	// with full pivoting.
	{
		var mat = new PortfolioAllocation.Matrix([[1, -1, 1], [-1, 1, -1], [1, -1, 2]]);
		var chol = PortfolioAllocation.Matrix.choleskyDecomposition(mat, {pivoting: "complete"}); 
		var L = chol.lowerTriangular;
		var D = chol.diagonal;
		var P = chol.permutation;
		
		// Ensure L is lower triangular
		// Ensure D has positive elements
		var lowerTriangular = true;
		var positiveDiagonalElements = true;
		for (var i = 1; i <= mat.nbRows; ++i) {
		  if (D.getValueAt(i,1) < 0) {
			  positiveDiagonalElements = false;
		  }
		  
		  for (var j = i+1; j <= mat.nbRows; ++j) {
			  if (L.getValueAt(i,j) != 0) {
				  lowerTriangular = false;
			  }
		  }
		}
		assert.equal(lowerTriangular, true, 'Cholesky decomposition with complete pivoting, semi-definite positive OK - Lower triangular');
		assert.equal(positiveDiagonalElements, true, 'Cholesky decomposition with complete pivoting, semi-definite positive OK - Positive diagonal elements');
				
		// Ensure P^t * A * P = L * Diag(D) * L^t
		var ldlt = PortfolioAllocation.Matrix.xy(PortfolioAllocation.Matrix.xy(L, PortfolioAllocation.Matrix.diagonal(D)), L.transpose());
		var patp = PortfolioAllocation.Matrix.xy(PortfolioAllocation.Matrix.txy(P, mat),P);
		assert.equal(PortfolioAllocation.Matrix.areEqual(ldlt, patp, 1e-12), true, 'Cholesky decomposition with complete pivoting, semi-definite positive OK - P^t * A * P = L * D * L^t');
  }
  
	// Test with static data the case of an indefinite matrix, which must fail
	{
		var mat = [[1,                    0.618421696258621  ,  0.4874298723751651,  -0.49079033902258856  , 0.11944609519358924 ,  -0.5578129586552424  , -0.4771156006785486 ,  -0.5891536415285011 ,  -0.3033661105868079 ,    0.346969569262485],
				[0.618421696258621,                     1  ,  0.5625061910498927 ,  -0.3066061449896621 ,  0.22036065346794634 ,  -0.2995526904104299 ,  -0.2665281748643263 ,  -0.4384779718501961  ,-0.04157807243431871 ,  0.36441921116789094],
				[0.4874298723751651 ,   0.5625061910498927  ,                   1,  -0.24898733591796773 ,  0.48207046028556727 ,  -0.3253014452697968 , -0.39490635735791035, -0.018501626749401468 , 0.034978192975079626,  -0.13605230462054563],
				[-0.49079033902258856,   -0.3066061449896621 , -0.24898733591796773 ,                    1,    0.3523200216615302 ,   0.5874187797438452,   0.35430037283032667,   0.28094692738663546  ,  0.5704634723157168,   -0.5160250884331637],
				[0.11944609519358924 ,  0.22036065346794634 ,  0.48207046028556727 ,   0.3523200216615302 ,                    1, -0.011807320173530558 ,  -0.1845706823452529 ,  0.48127181356733134 ,   0.5240171002774565 , -0.04905664181384413],
				[-0.5578129586552424 ,  -0.2995526904104299,   -0.3253014452697968 ,   0.5874187797438452, -0.011807320173530558,                     1 ,   0.9395942716104612 ,   0.3019431882153892 ,   0.5712314628087278 ,  -0.5702138190028966],
				[-0.4771156006785486,   -0.2665281748643263,  -0.39490635735791035 ,  0.35430037283032667,   -0.1845706823452529 ,   0.9395942716104612 ,                    1,    0.3594948044435324 ,  0.40258020819009377,  -0.23665358042668136],
				[-0.5891536415285011,   -0.4384779718501961, -0.018501626749401468 ,  0.28094692738663546,   0.48127181356733134 ,   0.3019431882153892,    0.3594948044435324 ,                    1,     0.307071908776885,  -0.27512543932175526],
				[-0.3033661105868079,  -0.04157807243431871,  0.034978192975079626 ,   0.5704634723157168,    0.5240171002774565 ,   0.5712314628087278,   0.40258020819009377,     0.307071908776885,                     1,  -0.12697231187851038],
				[0.346969569262485,   0.36441921116789094 , -0.13605230462054563  , -0.5160250884331637 , -0.04905664181384413  , -0.5702138190028966,  -0.23665358042668136,  -0.27512543932175526 , -0.12697231187851038  ,                   1]]
		var mat = PortfolioAllocation.Matrix(mat);
	  
		assert.throws(function() { 
			var chol = PortfolioAllocation.Matrix.choleskyDecomposition(mat, {pivoting: "complete"});
		},
		new Error('input matrix must be semi-definite positive'),
		"Cholesky decomposition with complete pivoting, semi-definite positive KO");
  }

  // TODO: Test using random data
  
  // TODO: Test error case
});


QUnit.test('Functions for correlation matrices polishing and testing', function(assert) {    
	// Test error cases
	{
		// Non square matrix
		var mat = PortfolioAllocation.Matrix.normrnd(10, 8);

		assert.throws(function() { mat.symmetrize(); },
						 new Error('matrix is not square: (10,8)'),
						 "Symmetrization - Non square matrix");
						 
		assert.throws(function() { mat.unitDiagonalize(); },
						 new Error('matrix is not square: (10,8)'),
						 "Unit diagonalization - Non square matrix");
		
        assert.equal(mat.isUnitDiagonal(), false, 'Unit diagonal testing - Non square matrix');
		
		// Non strictly unit diagonal matrix
		var mat = PortfolioAllocation.Matrix.identity(3);
		assert.equal(mat.isUnitDiagonal(), true, 'Unit diagonal testing - Identity');
		mat.setValueAt(1,1,1 + 1e-12);
		assert.equal(mat.isUnitDiagonal(), false, 'Unit diagonal testing - Identity perturbed');
		assert.equal(mat.isUnitDiagonal(1.01e-12), true, 'Unit diagonal testing - Identity perturbed #2');
		
		// Test for correlation matrix
		var mat = PortfolioAllocation.Matrix.normrnd(10, 8);
		assert.equal(mat.isCorrelationMatrix(), false, 'Correlation matrix testing - non square');
		mat = PortfolioAllocation.Matrix.normrnd(10, 10);
		assert.equal(mat.isCorrelationMatrix(), false, 'Correlation matrix testing - non symmetric');
		mat = PortfolioAllocation.Matrix.normrnd(10, 10).symmetrize();
		assert.equal(mat.isCorrelationMatrix(), false, 'Correlation matrix testing - non unit diagonal');
		mat = PortfolioAllocation.Matrix.normrnd(10, 10).symmetrize().unitDiagonalize();
		assert.equal(mat.isCorrelationMatrix(), false, 'Correlation matrix testing - non semi definite positive (unless bad luck)');
	}
	
	// Test using static data 
	{	  
		// Test matrix
		var mat = new PortfolioAllocation.Matrix([[1,2], [4,5]]);
		
		// Test symmetrization		
		var expectedMat = new PortfolioAllocation.Matrix([[1,3], [3,5]]);
		var diagMat = mat.symmetrize();		
		assert.deepEqual(diagMat.toArray(), expectedMat.toArray(), 'Functions for correlation matrices polishing and testing - D not sorted by default');
		assert.equal(diagMat.isUnitDiagonal(), false, 'Unit diagonal testing - Non unit diagonal matrix');
		
		// Test unit diagonalization		
		var expectedMat = new PortfolioAllocation.Matrix([[1,2], [4,1]]);
		var unitDiagMat = mat.unitDiagonalize();		
		assert.deepEqual(unitDiagMat.toArray(), expectedMat.toArray(), 'Functions for correlation matrices polishing and testing - D not sorted by default');
		assert.equal(unitDiagMat.isUnitDiagonal(), true, 'Unit diagonal testing - Unit diagonal matrix');
		
		// Test for correlation matrix
		var mat = new PortfolioAllocation.Matrix([[1,0.1], [0.1,1]]);
		assert.equal(mat.isCorrelationMatrix(), true, 'Correlation matrix testing - definite positive');
		
		var mat = new PortfolioAllocation.Matrix([[1,1], [1,1]]);
		assert.equal(mat.isCorrelationMatrix(), true, 'Correlation matrix testing - semi definite positive');
	}
});

QUnit.test('Functions for covariance matrices testing', function(assert) {    
	// Test error cases
	{
		// Non square matrix
		var mat = PortfolioAllocation.Matrix.normrnd(10, 8);
		assert.equal(mat.isCovarianceMatrix(), false, 'Covariance matrix testing - non square');
		assert.throws(function() { 
			mat.isCovarianceMatrix(undefined, "exception");
		},
		 new Error('not square'),
		 "Covariance matrix testing - non square");
		
		// Non symmetric
		mat = PortfolioAllocation.Matrix.normrnd(10, 10);
		assert.equal(mat.isCovarianceMatrix(), false, 'Covariance matrix testing - non symmetric');
		assert.throws(function() { 
			mat.isCovarianceMatrix(undefined, "exception");
		},
		 new Error('not symmetric'),
		 "Covariance matrix testing - non symmetric");
		
		// Non diagonal positive
		mat = PortfolioAllocation.Matrix.normrnd(10, 10).symmetrize().elemMap(function(i,j,val) { if (i == j) { return Math.min(-1e-16, val); } else { return val; }});
		assert.equal(mat.isCovarianceMatrix(), false, 'Covariance matrix testing - non positive diagonal');
		assert.throws(function() { 
			mat.isCovarianceMatrix(undefined, "exception");
		},
		 new Error('not positive diagonal'),
		 "Covariance matrix testing - non symmetric");
		
		// Non definite semi-positive
		mat = PortfolioAllocation.Matrix.normrnd(10, 10).symmetrize().elemMap(function(i,j,val) { if (i == j) { return Math.max(0, val); }  else { return val; }});
		assert.equal(mat.isCovarianceMatrix(), false, 'Covariance matrix testing - non semi definite positive (unless bad luck)');
		assert.throws(function() { 
			mat.isCovarianceMatrix(undefined, "exception");
		},
		 new Error('not positive semi-definite'),
		 "Covariance matrix testing - non symmetric");
	}
});


QUnit.test('Eigenvalues and eigenvectors computation', function(assert) {    
	// Test error cases
	{
		// Non square matrix
		assert.throws(function() { 
			var mat = PortfolioAllocation.Matrix.normrnd(10, 8);
			PortfolioAllocation.Matrix.eig(mat); 
		},
		 new Error('input matrix must be symmetric'),
		 "Eigenvalues and eigenvectors computation - Non square matrix");

		
		// Non symmetric matrix
		assert.throws(function() { 
			var mat = PortfolioAllocation.Matrix.normrnd(10, 10);
			PortfolioAllocation.Matrix.eig(mat); 
		},
		 new Error('input matrix must be symmetric'),
		 "Eigenvalues and eigenvectors computation - Non symmetric matrix");

	}
	
	// Test using static data for numerical errors
	// Reference: https://research.wmz.ninja/articles/2018/08/symmetrize-the-input-matrix-before-eigendecomposition.html
	{
		// Perturb the identity matrix so that it is non strictly symmetric
		var mat = PortfolioAllocation.Matrix.identity(3);
		mat.setValueAt(1,2,1e-15);
		
		// Computation of the eigenvalues and eigenvectors, which must NOT work without 
		// a tolerance value on numerical symmetry.
		//
		assert.throws(function() { 
			var jacobi = PortfolioAllocation.Matrix.eig(mat);
		},
		 new Error('input matrix must be symmetric'),
		 "Eigenvalues and eigenvectors computation - Numerically non symmetric matrix");
		
		//
		var jacobi = PortfolioAllocation.Matrix.eig(mat, {epsSymmetric: 1e-12});
		var V = jacobi[0];
		var D = jacobi[1];
		
		// Eigenvalues of identity matrix must all be numerically equal to 1
		var expectedEigenvalues = PortfolioAllocation.Matrix([1, 1, 1]);
		assert.equal(PortfolioAllocation.Matrix.areEqual(D, expectedEigenvalues, 1e-14), true, 'Eigenvalues and eigenvectors computation - perturbation of identity matrix #1');
		
		// Eigenvectors of identity matrix are any vectors, so that the matrix V must simply be orthogonal 
		assert.equal(PortfolioAllocation.Matrix.areEqual(PortfolioAllocation.Matrix.axty(1,V,V), 
		                                                 PortfolioAllocation.Matrix.identity(3), 1e-14), true, 'Eigenvalues and eigenvectors computation - perturbation of identity matrix #2');
	}
		
	// Test using static data 
	// Reference: The Jacobi Method for Real Symmetric Matrices, RUTISHAUSER
	{	  
		var mat = PortfolioAllocation.Matrix.fill(30, 30, function(i,j) { return Math.max(i,j);} );
	  
		// Computation of the eigenvalues and eigenvectors
		var jacobi = PortfolioAllocation.Matrix.eig(mat);
		var V = jacobi[0];
		var D = jacobi[1];
	  
		// Check that V is a square orthonormal matrix
		assert.equal(V.isSquare(), true, 'Eigenvalues and eigenvectors computation - V square');
		assert.equal(V.nbRows == 30, true, 'Eigenvalues and eigenvectors computation - V matrix of order 30');
		assert.equal(PortfolioAllocation.Matrix.areEqual(PortfolioAllocation.Matrix.txy(V, V), 
		                                                 PortfolioAllocation.Matrix.identity(30), 1e-14), true, 'Eigenvalues and eigenvectors computation - V orthonormal');	
	
		// Check that D is a vector
		assert.equal(D.isVector(), true, 'Eigenvalues and eigenvectors computation - D vector');
		assert.equal(D.nbRows == 30, true, 'Eigenvalues and eigenvectors computation - D vector of order 30');
		
		// Check that D is not sorted
		var isSorted = true;
		for (var i = 1; i < 30; ++i) {
			if ( D.data[i] > D.data[i-1] ) {
				isSorted = false;
				break;
			}
		}
		assert.equal(isSorted, false, 'Eigenvalues and eigenvectors computation - D not sorted by default');
		
		// Check that A = V * Diag(D) * V^t
		var mmat = PortfolioAllocation.Matrix.xy(V, PortfolioAllocation.Matrix.xy(PortfolioAllocation.Matrix.diagonal(D), V.transpose()));
		assert.equal(PortfolioAllocation.Matrix.areEqual(mat, mmat, 1e-13), true, 'Eigenvalues and eigenvectors computation - A equal to V*Diag(D)*V^t');

		// Check the computed eigenvalues are the same as the ones provided in the reference
		var expectedEigenvalues = [[1, 639.62943444], [2, -0.25068702021], [3, -0.25276325141], [16, -0.50027349839], [29, -24.077530173], [30, -114.51117646]];
		jacobi = PortfolioAllocation.Matrix.eig(mat, {sortedEigenvalues: true});
		DD = jacobi[1];
		
		for (var i = 0; i < expectedEigenvalues.length; ++i) {
			var expectedEigenValueIdx = expectedEigenvalues[i][0];
			var expectedEigenValue = expectedEigenvalues[i][1];
			assert.equal(Math.abs(expectedEigenValue - DD.data[expectedEigenValueIdx -1]) <= 1e-08, true, 'Eigenvalues and eigenvectors computation - Proper eigenvalues');
		}
		
		// Check that the new DD is sorted
		var isSorted = true;
		for (var i = 1; i < 30; ++i) {
			if ( DD.data[i] > DD.data[i-1] ) {
				isSorted = false;
				break;
			}
		}
		assert.equal(isSorted, true, 'Eigenvalues and eigenvectors computation - D sorted when requested');
	}
});



QUnit.test('Singular value decomposition', function(assert) {      
  // Test using static data, m = n case
  // Validated with Matlab
  {
	  var mat = new PortfolioAllocation.Matrix([[1,2,0], [1,1,1], [2,1,0]]);
	  
	  // Computation of a thin SVD decomposition
	  var svd = PortfolioAllocation.Matrix.svdDecomposition(mat);
	  var U = svd[0];
	  var S = svd[1];
	  var V = svd[2];

	  var expectedU = new PortfolioAllocation.Matrix([[  0.6279630301995544, 0.7071067811865475, -0.3250575836718681 ], 
													 [   0.459700843380983,                   0,  0.8880738339771154 ], 
													 [  0.6279630301995544,  -0.7071067811865475, -0.3250575836718681 ]]);
	  var expectedS = new PortfolioAllocation.Matrix([[  3.346065214951232,                  0,                  0 ],
													 [                  0,                  1,                  0 ],
													 [                  0,                  0, 0.8965754721680537 ]]);
	  var expectedV = new PortfolioAllocation.Matrix([[   0.7004017505420329,   -0.7071067811865475, -0.09714621885413724 ],
													 [   0.7004017505420329,  0.7071067811865475, -0.09714621885413724 ],
													 [  0.13738550023678578,                    0,   0.9905176547264003 ]]);
	 
	  assert.equal(PortfolioAllocation.Matrix.areEqual(U, expectedU), true, 'SVD decomposition, thin - #1 1/3');
	  assert.equal(PortfolioAllocation.Matrix.areEqual(S, expectedS), true, 'SVD decomposition, thin - #1 2/3');
	  assert.equal(PortfolioAllocation.Matrix.areEqual(V, expectedV), true, 'SVD decomposition, thin - #1 3/3');
  }

  // Test using static data, m > n case
  // Validated with Matlab
 {
	  var mat = new PortfolioAllocation.Matrix([[1,0,0,2], [0,4,0,0], [0,0,0,0], [0,2,0,0], [1,1,1,1]]);
	  
	  // Computation of a SVD decomposition
	  var svd = PortfolioAllocation.Matrix.svdDecomposition(mat, {svdForm: 'thin'});
	  var U = svd[0];
	  var S = svd[1];
	  var V = svd[2];

		var expectedU = new PortfolioAllocation.Matrix([[  0.04798852953924585,   0.8188713224171728,   -0.571967532606014,  -0.0479881585217683 ],
														[   0.8628011158778799, -0.16687220119644658,  -0.1665169748336121,  -0.8628011914856866 ],
														[                    0,                    0,                    0,                    0 ],
														[  0.43140055793893994, -0.08343610059822329, -0.08325848741680605,  -0.4314005957428433 ],
														[  0.25916190708226655,   0.5428092140960453,   0.7988699913064179, -0.25916166114085887 ]]);
		var expectedS = new PortfolioAllocation.Matrix([[       4.60450766417069,                      0,                      0,                      0 ],
														[                      0,     2.6436006852895395,                      0,                      0 ],
														[                      0,                      0,      0.899935879565907,                      0 ],
														[                      0,                      0,                      0, 1.3283077571435252e-19 ]]);
		var expectedV = new PortfolioAllocation.Matrix([[    0.06670646658090273,     0.5150855589084856,     0.2521318061125118,     -0.816496580927726 ],
														[     0.9931935876786804,   -0.11028586635967502,   -0.03746365005238518, -2.865163788807784e-20 ],
														[   0.056284390424387286,    0.20532950271821954,     0.8876965675507464,     0.4082482904638631 ],
														[    0.07712854273741816,     0.8248416150987519,   -0.38343295532572336,    0.40824829046386296 ]]);
														
	  assert.equal(PortfolioAllocation.Matrix.areEqual(U, expectedU), true, 'SVD decomposition, thin - #2 1/3');
	  assert.equal(PortfolioAllocation.Matrix.areEqual(S, expectedS), true, 'SVD decomposition, thin - #2 2/3');
	  assert.equal(PortfolioAllocation.Matrix.areEqual(V, expectedV), true, 'SVD decomposition, thin - #2 3/3');	  
  }

  // Test using static data, based on discussion: https://lists.gnu.org/archive/html/bug-gsl/2005-04/msg00008.html 
  {
	var mat1 = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,0,0], [0,0,0,1]]);
	var mat2 = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,0.1,0], [0,0,0,1]]);
	
    // Computation of a SVD decomposition of mat1
    var svd = PortfolioAllocation.Matrix.svdDecomposition(mat1);
    var S1 = svd[1];
	
	var expectedS1 = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,1,0], [0,0,0,0]]);
	assert.equal(PortfolioAllocation.Matrix.areEqual(S1, expectedS1), true, 'SVD decomposition, GSL error case - #1 1/2');
	
    // Computation of a SVD decomposition of mat2
    var svd = PortfolioAllocation.Matrix.svdDecomposition(mat2);
    var S2 = svd[1];
	
	var expectedS2 = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,1,0], [0,0,0,0.1]]);
	assert.equal(PortfolioAllocation.Matrix.areEqual(S2, expectedS2), true, 'SVD decomposition, GSL error case - #1 2/2');
  }
  
  // Unsupported case, m < n case
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4]]);
      assert.throws(function() { PortfolioAllocation.Matrix.svdDecomposition(mat); },
		                         new Error('matrix has more columns than rows: 3 v.s. 2'),
		                         "SVD decomposition, unsupported m < n case");
  }
  
    // TODO: Test using random data
});


QUnit.test('Nullspace computation', function(assert) {    
  // Test using static data, m < n case, nullspace dimension 0
  {
  	  var mat = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,1,0]]);
	  
	  // Computation of an orthogonal basis of Ker(A)
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat);
	  
	  var expectedNs = new PortfolioAllocation.Matrix([0,0,0,0]);
	  
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m < n - #1');
  }

  // Test using static data, m < n case, nullspace dimension 2
  // Validated with Wolfram Alpha
  {
  	  var mat = new PortfolioAllocation.Matrix([[1,0,0,0,2], [0,0,3,0,0], [0,0,0,0,0], [0,2,0,0,0]]);
	  
	  // Computation of an orthogonal basis of Ker(A)
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat);
	  
	  var expectedNs = new PortfolioAllocation.Matrix([[-0.8944271909999159, 0], 
													  [0, 0], 
													  [0, 0],
													  [0, -1], 
													  [0.4472135954999579, 0]]);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m < n - #2');
  }

  // Test using static data, m > n case, nullspace dimension 0
  {
  	  var mat = new PortfolioAllocation.Matrix([[1,0,0,0], [0,1,0,0], [0,0,1,0], [0,0,0,1], [1,1,1,1]]);
	  
	  // Computation of an orthogonal basis of Ker(A)
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat);
	  
	  var expectedNs = new PortfolioAllocation.Matrix([0,0,0,0]);
	  
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m > n - #1');
  }
  
  // Test using static data, m > n case, nullspace dimension 1
  // Validated with Wolfram Alpha
  {
  	  var mat = new PortfolioAllocation.Matrix([[1,0,0,2], [0,4,0,0], [0,0,0,0], [0,2,0,0], [1,1,1,1]]);
	  
	  // Computation of an orthogonal basis of Ker(A)
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat);
	  
	  var expectedNs = new PortfolioAllocation.Matrix([-0.816496580927726, -2.865163788807784e-20, 0.4082482904638631, 0.40824829046386296]);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m > n - #2');
  }
  
  // Test using static data, m = n case, nullspace of dimension 1 but of numerical dimension close to 0
  // Validated with Wolfram Alpha
  {
  	  var mat = new PortfolioAllocation.Matrix([[2,1],[-4,-2]]);
	  
	  // Computation of an orthogonal basis of Ker(A) with default tolerance
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat);
	  
	  var expectedNs = new PortfolioAllocation.Matrix([-0.4472135954999578, 0.8944271909999159]);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m == n - #1 1/2');
	  
	  // Computation of an orthogonal basis of Ker(A) with 1e-16 tolerance
	  var ns = PortfolioAllocation.Matrix.nullSpace(mat, {eps: 1e-16});
	  
	  var expectedNs = new PortfolioAllocation.Matrix([0, 0]);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(ns, expectedNs), true, 'Nullspace computation, m == n - #1 2/2');
  }
   
   // TODO: Test random data
});



QUnit.test('Determinant computation', function(assert) {    
  // Test using static data  
  var mat = new PortfolioAllocation.Matrix([[-2,2,-3], [-1,1,3], [2,0,-1]]);
  var expectedValue = 18;

  assert.equal(Math.abs(mat.determinant() - expectedValue) <= 1e-16, true, 'Determinant computation');
  
  // TODO: Test random data
});


QUnit.test('Trace computation', function(assert) {    
  // Test using static data  
  var mat = new PortfolioAllocation.Matrix([[-2,2,-3], [-1,1,3], [2,0,-1]]);
  var expectedValue = -2;

  assert.equal(Math.abs(mat.trace() - expectedValue) <= 1e-16, true, 'Trace computation');
  
  // TODO: Test random data
});


QUnit.test('Matrix copy', function(assert) {    
  // Test using static data
  {
    // Original matrix to be copied, and expected copy
    var originalMat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
    var expectedCopyMat = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
    
    // First copy: standard one
    var copyMat = PortfolioAllocation.Matrix.copy(originalMat);
    
    // Second copy: using an already existing matrix
    var outputMat = PortfolioAllocation.Matrix.zeros(originalMat.nbRows, originalMat.nbColumns);
    var outputCopyMat = PortfolioAllocation.Matrix.copy(originalMat, outputMat);
    
    // Alteration of the original matrix to check that the copy is a real copy and not a reference
    originalMat.setValueAt(1, 1, 0);
    var originalUpdatedMat = new PortfolioAllocation.Matrix([[0,2,3], [4,5,6]]);
    assert.equal(PortfolioAllocation.Matrix.areEqual(originalMat, originalUpdatedMat), true, 'Matrix copy #1/0');
    
    // Tests of the method behaviour
    assert.equal(PortfolioAllocation.Matrix.areEqual(copyMat, expectedCopyMat), true, 'Matrix copy #1/3');
    assert.equal(PortfolioAllocation.Matrix.areEqual(outputMat, outputCopyMat), true, 'Matrix copy #1/3');
    assert.equal(PortfolioAllocation.Matrix.areEqual(outputCopyMat, expectedCopyMat), true, 'Matrix copy #1/2');
  }
  
  // TODO: Test random data
});


QUnit.test('Random matrix generation', function(assert) {     
  // Test normal generation using random data
  {
	  // Define the number of tests and the min/max dimensions
	  var nbTests = 10;
	  var minDimension = 1;
	  var maxDimension = 100;
	  
	  //
	  var size = true
	  for (var i = 0; i < nbTests; ++i) {
		// Generate the dimensions
		var m = Math.floor(Math.random()*(maxDimension-minDimension+1) + minDimension)
		var n = Math.floor(Math.random()*(maxDimension-minDimension+1) + minDimension)
		
		// Generate the matrix
		var mat = PortfolioAllocation.Matrix.normrnd(m, n);

		// Check the size of the matrix
		assert.equal(mat.nbRows == m, true, 'Random matrix generation, positive normal - Test #' + i + ', number of rows');
		assert.equal(mat.nbColumns == n, true, 'Random matrix generation, positive normal - Test #' + i + ', number of columns');
	  }
  }
});



QUnit.test('Random orthogonal matrix generation', function(assert) {    
  // Test using random data that the generated matrices O are orthogonal:
  // - O must be n by n
  // - O must be such that O*O^t = O^t * O = Identity(n)
  {
      var nbTests = 10;
	  
	  for (var i = 0; i < nbTests; ++i) {
		  // Generate a random dimension
		  var n_max = 200;
		  var n_min = 2;
		  var n = Math.floor(Math.random()*(n_max-n_min+1) + n_min);
		  
		  // Generate a random matrix
		  var mat = PortfolioAllocation.Matrix.randomOrthogonal(n);
		  
		  // Check that the matrix is square
		  assert.equal(mat.nbRows == n && mat.nbColumns == n, true, 'Random orthogonal matrix generation, square - Test #' + i);
		  
		  // Check that the matrix is orthogonal
		  var mat_p_t_mat = PortfolioAllocation.Matrix.axty(1,mat,mat);
		  assert.equal(PortfolioAllocation.Matrix.areEqual(mat_p_t_mat, PortfolioAllocation.Matrix.identity(n), 1e-12), true,'Random orthogonal matrix generation, orthogonal 1 - Test #' + i);
		  
		  var t_mat_p_mat = PortfolioAllocation.Matrix.atxy(1,mat,mat);
		  assert.equal(PortfolioAllocation.Matrix.areEqual(t_mat_p_mat, PortfolioAllocation.Matrix.identity(n), 1e-12), true,'Random orthogonal matrix generation, orthogonal 2 - Test #' + i);
	  }
  }
});


QUnit.test('Random correlation matrix generation', function(assert) {    
	// Test using random data that the generated matrices C are correlation matrix:
	// - C must be n by n
	// - C must be symmetric positive semidefinite, with unit diagonal
	{
		var nbTests = 10;
		
		for (var l = 0; l < nbTests; ++l) {
			// Generate a random dimension
			var n_max = 200;
			var n_min = 2;
			var n = Math.floor(Math.random()*(n_max-n_min+1) + n_min);
			
			// Generate a random matrix
			var mat = PortfolioAllocation.Matrix.randomCorrelation(n);

			// Check that the matrix is square
			assert.equal(mat.nbRows == n && mat.nbColumns == n, true, 'Random correlation matrix generation, square - Test #' + l);
			
			// Check that the matrix is truly symmetric
			assert.equal(mat.isSymmetric(), true, 'Random correlation matrix generation, symmetric - Test #' + l);
			
			// Check that the matrix has true unit diagonal
			var unit_diagonal = true;
			for (var i = 1; i <= n; ++i) {
				if (mat.getValue(i,i) != 1) {
					unit_diagonal = false;
				}
			}
			assert.equal(unit_diagonal, true, 'Random correlation matrix generation, unit diagonal - Test #' + l);
			
			// Check that the matrix is positive semidefinite
			// Check that all the eigenvalues are >= 0, which is implied by the fact that
			// the lowest eigenvalue is >= 0
			var jacobi = PortfolioAllocation.Matrix.eig(mat, {sortedEigenvalues: true});
			var D = jacobi[1];
			assert.equal(D.data[n-1] >= 0, true, 'Random correlation matrix generation, semi-positive definite - Test #' + l);
			
			// Double check with the usage of the dedicated method
			assert.equal(mat.isCorrelationMatrix(), true, 'Random correlation matrix generation, all inclusive - Test #' + l);
		}
	}
	
	// Test using static data conditions on eigenvalues when they are provided
	{
		// The number of eigenvalues must equal the dimension
		assert.throws(function() { PortfolioAllocation.Matrix.randomCorrelation(3, {lambda: [1, 2]}) },
		                           new Error('number of input eigenvalues not equal to 3, but to 2'),
		                           "Random correlation matrix generation, incorrect number of provided eigenvalues");
		
		// The eigenvalues must sum to the dimension
		assert.throws(function() { PortfolioAllocation.Matrix.randomCorrelation(3, {lambda: [1, 2, 3]}) },
		                           new Error('input eigenvalues not summing to 3'),
		                           "Random correlation matrix generation, provided eigenvalues not summing to 1");
	}
	
	// Test using random data that when eigenvalues are provided, they
	// correspond to the eigenvalues of the generated matrix
	{
		var nbTests = 10;
		
		for (var l = 0; l < nbTests; ++l) {
			// Generate a random dimension
			var n_max = 100;
			var n_min = 2;
			var n = Math.floor(Math.random()*(n_max-n_min+1) + n_min);
			
			// Generate random eigenvalues
			//
			// For ease of testing, let's take 1..n, and then normalize them so that they sum to n
			var lambdas = new Array(n);
			var sum_lambas = n*(n+1)/2;
			var gg=0;
			for (var i = 1; i <= n; ++i) {
				lambdas[i-1] = n * i/sum_lambas;
			}
			
			// Generate a random matrix
			var mat = PortfolioAllocation.Matrix.randomCorrelation(n, {lambda: lambdas});

			// Check that the matrix is square
			assert.equal(mat.nbRows == n && mat.nbColumns == n, true, 'Random correlation matrix generation, provided eigenvalues, square - Test #' + l);
			
			// Check that the matrix is numerically symmetric
			assert.equal(mat.isSymmetric(1e-14), true, 'Random correlation matrix generation, provided eigenvalues, symmetric - Test #' + l);
			
			// Check that the matrix has true unit diagonal
			var unit_diagonal = true;
			for (var i = 1; i <= n; ++i) {
				if (mat.getValue(i,i) != 1) {
					unit_diagonal = false;
				}
			}
			assert.equal(unit_diagonal, true, 'Random correlation matrix generation, provided eigenvalues, unit diagonal - Test #' + l);
			
			// Check that the eigenvalues of the matrix are the provided eigenvalues
			//
			// Note: lambda is an eigenvalue of matrix A iif det(A - lambda*Id) == 0
			var providedEigenvaluesMatching = true;
			var eye = PortfolioAllocation.Matrix.identity(n);
			for (var i = 0; i < n; ++i) {
				if ( Math.abs(PortfolioAllocation.Matrix.axpby(1, mat, -lambdas[i], eye).determinant()) > 1e-14) {
					providedEigenvaluesMatching = false;
					break;
				}
			}
			assert.equal(providedEigenvaluesMatching, true, 'Random correlation matrix generation, provided eigenvalues matching - Test #' + l);
		}
	}
});

QUnit.test('Matrix AXPBY', function(assert) {    
  // Test using static data
  {
    // 
    var mat1 = new PortfolioAllocation.Matrix([[1,2,3], [4,5,6]]);
	var a = -1;
    var mat2 = new PortfolioAllocation.Matrix([[7,8,9], [10,11,12]]);
	var b = 1;
	var expectedMat = new PortfolioAllocation.Matrix([[6,6,6], [6,6,6]]);
	
	// First computation: standard one
	var computedMat = PortfolioAllocation.Matrix.axpby(a, mat1, b, mat2);
	
	// Second computation: using an already existing matrix
    var outputMat = PortfolioAllocation.Matrix.zeros(mat1.nbRows, mat1.nbColumns);
	var outputComputedMat = PortfolioAllocation.Matrix.axpby(a, mat1, b, mat2, outputMat);
	
	// Tests of the method behaviour
	assert.equal(PortfolioAllocation.Matrix.areEqual(computedMat, expectedMat), true, 'Matrix axpby #1/1');
	assert.equal(PortfolioAllocation.Matrix.areEqual(outputMat, outputComputedMat), true, 'Matrix axpby #1/2');
	assert.equal(PortfolioAllocation.Matrix.areEqual(outputComputedMat, expectedMat), true, 'Matrix axpby #1/3');
	
	// Test that incrementing in place a matrix is possible
	var mat3 = new PortfolioAllocation.Matrix.ones(3, 3);
	var mat4 = PortfolioAllocation.Matrix.axpby(1, mat3, 2, mat3, mat3);
	var expectedMat = new PortfolioAllocation.Matrix([[3,3,3], [3,3,3], [3,3,3]]);
	assert.equal(PortfolioAllocation.Matrix.areEqual(mat3, expectedMat), true, 'Matrix axpby #2/1');
	assert.equal(mat3 === mat4, true, 'Matrix axpby #2/2');
  }
  
  // TODO: Test random data
});


QUnit.test('Matrix norms computation', function(assert) {    
  // Unsupported norm
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,2,8]]);
      assert.throws(function() { mat.matrixNorm(2) },
		                         new Error('unsupported matrix norm: 2'),
		                         "Matrix norm computation - unsupported norm");
  }
		
  // Test using static data  
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,2,8]]);
      var expectedNorm1 = 19;
      var expectedNormInf = 15;
      var expectedNormFrobenius = Math.sqrt(3*3 + 5*5 + 7*7 + 2*2 + 6*6 + 4*4 + 2*2 + 8*8);

      assert.equal(mat.matrixNorm('one'), expectedNorm1, 'Matrix 1-norm computation #1');
      assert.equal(mat.matrixNorm('infinity'), expectedNormInf, 'Matrix infinity-norm computation #1');
      assert.equal(mat.matrixNorm('frobenius'), expectedNormFrobenius, 'Matrix Frobenius-norm computation #1');
  }

  // Test using static data  
  {
      var mat = new PortfolioAllocation.Matrix([[0,0], [0,0]]);
      var expectedNorm1 = 0;
      var expectedNormInf = 0;
      var expectedNormFrobenius = 0;

      assert.equal(mat.matrixNorm('one'), expectedNorm1, 'Matrix 1-norm computation #2');
      assert.equal(mat.matrixNorm('infinity'), expectedNormInf, 'Matrix infinity-norm computation #2');
      assert.equal(mat.matrixNorm('frobenius'), expectedNormFrobenius, 'Matrix Frobenius-norm computation #2');
  }
  
  // TODO: Test random data
});

QUnit.test('Matrix minimum and maximum values computation', function(assert) {    
  // Test using static data  
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,2,8]]);
      var expectedMinVal = -3;
	  var expectedMaxVal = 8;

      assert.equal(mat.min(), expectedMinVal, 'Matrix minimum computation #1');
	  assert.equal(mat.max(), expectedMaxVal, 'Matrix maximum computation #1');
  }
});

QUnit.test('Matrix vector norms computation', function(assert) {    
  // Unsupported norm
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,2,8]]);
      assert.throws(function() { mat.vectorNorm(2) },
		                         new Error('unsupported vector norm: 2'),
		                         "Matrix vector norm computation - unsupported norm");
  }
		
  // Test using static data  
  {
      var mat = new PortfolioAllocation.Matrix([[-1,2], [3,-4]]);
      var expectedNorm1 = 10;
      var expectedNormInf = 4;
      var expectedNorm2 = Math.sqrt(30);

      assert.equal(mat.vectorNorm('one'), expectedNorm1, 'Matrix vector 1-norm computation');
      assert.equal(mat.vectorNorm('infinity'), expectedNormInf, 'Matrix vector infinity-norm computation');
      assert.equal(mat.vectorNorm('two'), expectedNorm2, 'Matrix vector 2-norm computation');
  }
  
  // TODO: Test random data
});


QUnit.test('Matrix row/columns norms computation', function(assert) {    
  // Unsupported norm
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,2,8]]);
      assert.throws(function() { mat.vectorNorm(2, 'row', 1) },
		                         new Error('unsupported vector norm: 2'),
		                         "Matrix row norm computation - unsupported norm");
  }
		
  // Test using static data  
  {
      var mat = new PortfolioAllocation.Matrix([[-3,5,7], [2,6,4], [0,0,0]]);
      var expectedRowsNorm1 = [15, 12, 0];
      var expectedRowsNormInf = [7, 6, 0];
      var expectedRowsNorm2 = [Math.sqrt(3*3 + 5*5 + 7*7), Math.sqrt(2*2 + 6*6 +4*4), 0];
	  
	  for(var i = 0; i < mat.nbRows; ++i) {
		assert.equal(mat.vectorNorm('one', 'row', i+1), expectedRowsNorm1[i], 'Matrix row 1-norm computation');
		assert.equal(mat.vectorNorm('infinity', 'row', i+1), expectedRowsNormInf[i], 'Matrix row infinity-norm computation');
		assert.equal(Math.abs(mat.vectorNorm('two', 'row', i+1) - expectedRowsNorm2[i]) <= 1e-14, true, 'Matrix row 2-norm computation');
	  }
	  
	  
      var expectedColumnsNorm1 = [5, 11, 11];
      var expectedColumnsNormInf = [3, 6, 7];
      var expectedColumnsNorm2 = [Math.sqrt(3*3 + 2*2), Math.sqrt(5*5 + 6*6), Math.sqrt(7*7 + 4*4)];
	  
	  for(var j = 0; j < mat.nbColumns; ++j) {
		assert.equal(mat.vectorNorm('one', 'column', j+1), expectedColumnsNorm1[j], 'Matrix column 1-norm computation');
		assert.equal(mat.vectorNorm('infinity', 'column', j+1), expectedColumnsNormInf[j], 'Matrix column infinity-norm computation');
		assert.equal(Math.abs(mat.vectorNorm('two', 'column', j+1) - expectedColumnsNorm2[j]) <= 1e-14, true, 'Matrix column 2-norm computation');
	  }
  }
  
  // TODO: Test random data
});


QUnit.test('Linear system solver - Systems solve via back and forward substitution', function(assert) {    
  // Test using static data  
  {
      var A = new PortfolioAllocation.Matrix([[1, 2, 1, -1], [0, -4, 1, 7], [0, 0, -2, 1], [0, 0, 0, -1]]);
	  var b = new PortfolioAllocation.Matrix([5, 1, 1, 3]);
      var expectedX = new PortfolioAllocation.Matrix([16, -6, -2, -3]);
	  
	  var x = PortfolioAllocation.Matrix.linsolveBackSubstitution(A, b);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-14), true, 'Upper triangular linear system solve - Back substitution algorithm #1');
  }
  
  // Test using static data  
  {
      var A = new PortfolioAllocation.Matrix([[2, 0, 0], [1, 5, 0], [7, 9, 8]]);
	  var b = new PortfolioAllocation.Matrix([6, 2, 5]);
      var expectedX = new PortfolioAllocation.Matrix([3, -0.2, -1.775]);
	  
	  var x = PortfolioAllocation.Matrix.linsolveForwardSubstitution(A, b);
	  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-14), true, 'Lower triangular linear system solve - Forward substitution algorithm #1');
  }
  
  // TODO: Test random data
});


QUnit.test('Linear system solver - Extended Kaczmarz algorithm', function(assert) {    
	// Test the two different Kaczmarz algorithms
	var randomizedValues = [false, true];
	for (var idx = 0; idx < randomizedValues.length; ++idx) {		
	  var randomizedValue = randomizedValues[idx];
	  
	  // Limit case: null matrix; in this case, the least square solution is null as well
	  {
		  var A = new PortfolioAllocation.Matrix([[0,0,0], [0,0,0], [0, 0, 0]]);
		  var b = new PortfolioAllocation.Matrix([1,2,3]);
		  var expectedX = new PortfolioAllocation.Matrix([0, 0, 0]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-14), true, 'Linear system solve - Extended Kaczmarz algorithm #1');
	  }
	  
	  // Test using static data
	  {
		  // Source: https://en.wikipedia.org/wiki/System_of_linear_equations
		  var A = new PortfolioAllocation.Matrix([[3,2,-1], [2,-2,4], [-1, 0.5, -1]]);
		  var b = new PortfolioAllocation.Matrix([1,-2,0]);
		  var expectedX = new PortfolioAllocation.Matrix([1, -2, -2]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-8), true, 'Linear system solve - Extended Kaczmarz algorithm #2');
	  }
	  
	  // Reference: Tanabe, K.: An algorithm for the constrained maximization in nonlinear programming, Research Memorandum No. 31, The Institute of Statistical Mathematics, 1969.
	  // Test using static data
	  {
		// Problem 1
		var A = new PortfolioAllocation.Matrix([[-3.2, 2.9, 1.6, 0.1], [0.0, -1.1, 2.3, 1.0], [5.1, 4.8, 0.2, 4.9], [2.0, 1.1, 1.9, -2.9]]);
		var b = new PortfolioAllocation.Matrix([1.4, 2.2, 15.0, 2.1]);
		var expectedX = new PortfolioAllocation.Matrix([1, 1, 1, 1]);
		
		var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-10), true, 'Linear system solve - Extended Kaczmarz algorithm #3');
		
		
		// Problem 4
		var A = new PortfolioAllocation.Matrix([[1, 3, 2, -1], [1, 2, -1, -2], [1, -1, 	2, 3], [2, 1, 1, 1], [5, 5, 4, 1], [4, -1, 5, 7]]);
		var b = new PortfolioAllocation.Matrix([5, 0, 5, 5, 15, 15]);
		// To be noted that the true solution in the reference is [1, 1, 1, 1], with A*x - b == 0 and ||x||_2 == 2
		// The expected solution below verifies ||A*x - b||_inf ~= 1e-13, but ||x||_2 ~= 1.96, which is less than 2, hence why it is preferred to the true solution.
		var expectedX = new PortfolioAllocation.Matrix([1.1538461538462281, 0.7692307692307582, 1.153846153846079, 0.7692307692307796]);

		var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-10), true, 'Linear system solve - Extended Kaczmarz algorithm #4');
		
		
		// Problem 6
		var A = PortfolioAllocation.Matrix.fill(84, 84, function(i,j) { if (i == j) { return 6; } else if (j == i+1) { return 1; } else if (i == j+1) { return 8; } else { return 0;} });
		var b = PortfolioAllocation.Matrix.fill(84, 1, function(i,j) { if (i == 1) { return 7; } else if (i == 84) { return 14; } else { return 15;} });
		// To be noted that the true solution in the reference is [1, ..., 1], with A*x - b == 0 and ||x||_2 > ~9.16
		// The expected solution below verifies ||A*x - b||_inf ~= 1e-13, but ||x||_2 > ~9.15, which is less than 9.16, hence why it is preferred to the true solution.
		// To also be noted that the solution in the reference is almost the same as the expected solution below.
		var expectedX = new PortfolioAllocation.Matrix([1, 0.9999999999999999, 1.0000000000000002, 0.9999999999999996, 1.0000000000000004, 0.9999999999999996, 1.0000000000000004, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999997, 1.0000000000000002, 0.9999999999999999,                  1,                  1, 0.9999999999999999, 1.0000000000000002, 0.9999999999999998, 1.0000000000000002, 0.9999999999999992,  1.000000000000001, 0.9999999999999987,  1.000000000000002, 0.9999999999999974, 1.0000000000000024, 0.9999999999999983, 0.9999999999999991,  1.000000000000007, 0.9999999999999785, 1.0000000000000522, 0.9999999999998842, 1.0000000000002456, 0.9999999999994922,  1.000000000001035, 0.9999999999979075, 1.0000000000042115,  0.999999999991547, 1.0000000000169387, 0.9999999999660868, 1.0000000000678653, 0.9999999998642277, 1.0000000002715892,  0.999999999456776, 1.0000000010864951, 0.9999999978269619, 1.0000000043461248,  0.999999991307702, 1.0000000173846435,  0.999999965230667, 1.0000000695387083,  0.999999860922549, 1.0000002781549087, 0.9999994436902785, 1.0000011126189468, 0.9999977747641977,  1.000004450463144, 0.9999910991076416, 1.0000178016489207, 0.9999643972454112,  1.000071203336108, 0.9998576020201143, 1.0002847611904067, 0.9994306166966817, 1.0011382102966262, 0.9977258046468135, 1.0045394897460913, 0.9909566243489603, 1.0179443359374987, 0.9646809895833339, 1.0683593749999998, 0.8723958333333336, 1.2187499999999998, 0.7083333333333335]);
		
		var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-10), true, 'Linear system solve - Extended Kaczmarz algorithm #5');
	  }
	  
	  // Test using static data
	  // Source: http://www.math.usm.edu/lambers/mat419/lecture15.pdf
	  {
		  var A = new PortfolioAllocation.Matrix([[1,1,1], [-1,-1,1]]);
		  var b = new PortfolioAllocation.Matrix([1,0]);
		  var expectedX = new PortfolioAllocation.Matrix([1/4, 1/4, 1/2]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-12), true, 'Linear system solve (underdetermined) - Extended Kaczmarz algorithm #6');
	  }
	  
	  // Test using static data
	  // Source: http://math.oit.edu/~watermang/math_341/341_ch8/F13_341_book_sec_8-5.pdf
	  {
		  var A = new PortfolioAllocation.Matrix([[1.3,0.6], [4.7, 1.5], [3.1, 5.2]]);
		  var b = new PortfolioAllocation.Matrix([3.3, 13.5, -0.1]);
		  var expectedX = new PortfolioAllocation.Matrix([3.5526, -2.1374]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-04), true, 'Linear system solve (overdetermined) - Extended Kaczmarz algorithm #7');
	  }
	  
	  // Test using static data
	  // Source: Internet
	  {
		  var A = new PortfolioAllocation.Matrix([[1, 0, -1, 2], [1, 1, 1, -1], [0, -1, -2, 3], [5, 2, -1, 4], [-1, 2, 5, -8]]);
		  var b = new PortfolioAllocation.Matrix([-1, 2, -3, 1, 7]);
		  var expectedX = new PortfolioAllocation.Matrix([0.5, 0.5, 0.5, -0.5]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-10), true, 'Linear system solve (overdetermined, rank deficient) - Extended Kaczmarz algorithm #8');
	  }
	  
	  // Test using static data
	  // Source: Internet
	  {
		  var A = new PortfolioAllocation.Matrix([[1,3,-2,3,8,0], [-3,0,0,1,9,4], [-2,3,-2,4,17,4]]);
		  var b = new PortfolioAllocation.Matrix([1, 2, -3]);
		  var expectedX = new PortfolioAllocation.Matrix([-0.0783, -0.0778, 0.0519, -0.0604, -0.0504, 0.0698]);

		  var x = PortfolioAllocation.Matrix.linsolveExtendedKaczmarz(A, b, {randomized: randomizedValue});
		  assert.equal(PortfolioAllocation.Matrix.areEqual(x, expectedX, 1e-04), true, 'Linear system solve (overdetermined, rank deficient) - Extended Kaczmarz algorithm #9');
	  }
	}
	
	  
  /*
  	var A = new PortfolioAllocation.Matrix([[    0.632810907566168,  0.07458558260068052,  0.18972932155688532,    0.368496003408602 ],
											[  0.07458558260068052, 0.014210261176971236,  0.02239443845104238,   0.0379808829726669 ],
											[  0.18972932155688532,  0.02239443845104238,  0.05694725645543196,    0.110387626650411 ],
											[    0.368496003408602,   0.0379808829726669,    0.110387626650411,  0.22012749378552407 ]]);
	var b = new PortfolioAllocation.Matrix([ 0.0034247797381810052, 0.11594559772023327,  0.002703559429703406, -0.11522437741175573 ]);
	// cond(2-norm) =   14319829869135788.0000
	 
	var x = PortfolioAllocation.Matrix.linsolveRandomizedExtendedKaczmarz(A, b, {maxIter: -1});
	*/
  // TODO: Test random data

  // TODO: Test random data
});